// include para usar el BLE
#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>
// include para usar librerías y codigo en C
#include <stdio.h>
#include <string.h>
#include <iostream>
#include <fstream>
//libreria para usar el led RGB
#include <analogWrite.h>
// include para usar la tarjeta de memoria SD, y tener las conexiones correctas
//#include <SPI.h>      //esta libreria no hace falta incluirla aqui, pues ya lo hace de por sí la SD.h
#include <SD.h>

#include "defines.h"

BLEServer *mi_servidor;
BLEService *servicio_transferencia;
BLECharacteristic *caracteristica_envio_estado_actual;
BLECharacteristic *caracteristica_envio_tiempo;
BLECharacteristic *caracteristica_envio_nombre_ciclo;
BLECharacteristic *caracteristica_recibo_ciclo;
BLECharacteristic *caracteristica_recibo_nombre_tiempo;
bool deviceConnected = false; //inicializamos la conexion del dispositivo como "falso"
int txValue=0;

#define SERVICE_UUID                                 "6E400001-B5A3-F393-E0A9-E50E24DCCA9E"

#define CHARACTERISTIC_UUID_ESTADO                   "6E400003-B5A3-F393-E0A9-ECECECEBBBBB"    // EC --> caracteristica ENVIAR estado
#define CHARACTERISTIC_UUID_TIEMPO                   "6E400003-B5A3-F393-E0A9-ACACACABBBBB"    // AC --> caracteristica ENVIAR tiempo
#define CHARACTERISTIC_UUID_CICLO                    "6E400003-B5A3-F393-E0A9-CCCCCCCBBBBB"    // CC --> caracteristica ENVIAR ciclo

#define CHARACTERISTIC_UUID_RECIBE_CICLO             "6E400003-B5A3-F393-E0A9-FCCFCCFCCBBB"    // FCC --> caracteristica RECIBIR ciclo
#define CHARACTERISTIC_UUID_RECIBE_NOMBRE_TIEMPO     "6E400003-B5A3-F393-E0A9-FECFECFECBBB"    // FEC --> caracteristica RECIBIR nombre del ciclo de trabajo y tiempo entre estados




//led RGB previsto para indicar el estado del dsipositivo esclavo
#define rgb_verde 2     // Este es el orden correcto. En la placa de pruebas V2.0 me equivoqué con los colores de los cables azul y verde (están intercambiados)
#define rgb_rojo 15
#define rgb_azul 4

// conexiones del modulo lector de tarjetas SD
#define CS 5
/*          Los siguientes " #define " simplemente están dentro de los comentarios para tener en cuenta que esos pines están en uso
#define SCK 18
#define MOSI 23
#define MISO 19
*/

char panel_led[8]={led0,led1,led2,led3,led4,led5,led6,led7};
char estado[8];
int led_rgb[3] = {rgb_verde,rgb_rojo,rgb_azul};
int delay_programa = 0;
bool ciclo_actualizado = 1;                         //Esta variable será usada para indicar la actualizacion del los ficheros de trabajo a usar referidos al ciclo de trabajo
bool tiempo_actualizado = 1;                        //Esta variable será usada para indicar la actualizacion del los ficheros de trabajo a usar referidos al tiempo entre estados
bool nombre_actualizado = 1;                        //Esta variable será usada para indicar la actualizacion del los ficheros de trabajo a usar referidos al nombre del ciclo de trabajo
bool escribiendo_ciclo = 0;                         //Variable para indicar que NO se está procediendo a la escritura del fichero correspondiente al ciclo de trabajo y puede accederse a el
bool escribiendo_tiempo_nombre = 0;                 //Variable para indicar que NO se está procediendo a la escritura del fichero correspondiente al tiempo y nombre del ciclo de trabajo y puede accederse a el


File archivo;                                       //variable con que accederemos a los archivos a usar en cada momento
File nuevo;                                         //variable que usaré para escribir los nuevos ficheros según reciba los datos desde el maestro
String archivo_ciclo_actual;                        //variable que nos dará la ruta al archivo del ciclo actual de trabajo
String archivo_tiempo_actual;                       //variable que nos dará la ruta al archivo en que se fija el tiempo entre estados
String archivo_nombre_ciclo_actual;                 //variable que nos dará la ruta al archivo que almacena el nombre del ciclo que estamos ejecutando en el esclavo

int color = 0;                                      //variable con que definimos el color del led RGB que se usa para indicar el estado de la conexión BLE


//*************************************************************************************************************************************************************************************************************************************
//**********************************************************     MODULOS  REFERENTES  A  LA  ESCRITURA  EN  LAS  CARACTERISTICAS      *************************************************************************************************
//*************************************************************************************************************************************************************************************************************************************

/*Este modulo servirá para enviar al dispositivo maestro el estado de la secuencia en que se encuentra el modulo esclavo. 
Dado que le queremos enviar un estado que ya está como char array, lo hacemos directamente */
void escribir_estado()                                                                                                                                                                                             //  MODULO  ACABADO
{
    caracteristica_envio_estado_actual -> setValue(estado);
    caracteristica_envio_estado_actual -> notify();

    //Muestro por pantalla y meto un delay para que haya tiempo suficiente como para ver las diferentes lecturas
    //Serial.println("Valor del string: " + String(estado));
    //delay(1000);
}
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

/*Este modulo servirá para enviar al dispositivo maestro el nombre del ciclo en curso actual que está ejecutando el modulo esclavo. 
Dado que le queremos enviar el nombre de un ciclo de trabajo, debemos pasar de string a char array */
void escribir_ciclo_en_curso(String ciclo_en_curso)                                                                                                                                                                 //  MODULO  ACABADO
{
    char nombre_ciclo[ciclo_en_curso.length() + 1];// creo un array char con el tamaño de el string recibido, más 1 para el fin del mismo

    ciclo_en_curso.toCharArray(nombre_ciclo,ciclo_en_curso.length()+1);// paso el valor de string a char array

    caracteristica_envio_nombre_ciclo -> setValue((char*)&nombre_ciclo);  // escribo el valor en mi característica
    caracteristica_envio_nombre_ciclo -> notify();  //notifico el valor que acabo de escribir

    //Muestro por pantalla y meto un delay para que haya tiempo suficiente como para ver las diferentes lecturas
    //Serial.println("Nombre del ciclo en curso: " + String(ciclo_en_curso));
    //delay(1000);
}
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

/*Este modulo servirá para enviar al dispositivo maestro el nombre del ciclo en curso actual que está ejecutando el modulo esclavo. 
Dado que le queremos enviar el tiempo entre estados, debemos pasar de int a char array */
void escribir_tiempo_ciclo(int tiempo_entre_estados)                                                                                                                                                                //  MODULO  ACABADO
{
    char tiempo[5];// creo un array char con 4 espacios, pues esto ya nos basta para un tiempo de hasta 9999 segundos, lo que suponen 2 horas y 26 minutos como tiempo entre 2 estados, más que suficiente

    dtostrf(tiempo_entre_estados,1,0,tiempo); // paso el valor de int a char array

    caracteristica_envio_tiempo -> setValue((char*)&tiempo);  // escribo el valor en mi característica
    caracteristica_envio_tiempo -> notify();  //notifico el valor que acabo de escribir

    //Muestro por pantalla y meto un delay para que haya tiempo suficiente como para ver las diferentes lecturas
    //Serial.println("Nombre del ciclo en curso: " + tiempo_entre_estados);
    //sdelay(1000);
}

//*************************************************************************************************************************************************************************************************************************************
//*********************************************************     MODULOS  REFERENTES  A  LA  CARGA  Y  USO DE  DATOS  DESDE  EL  FICH.       *******************************************************************************************
//*************************************************************************************************************************************************************************************************************************************

void cargar_estado()                                                                                                                                                                                                //  MODULO  ACABADO
{
    char lectura;
    for (int i = 0 ; i <= 7; i++)      //este bucle es el que se encarga de gesionar los diferentes led del panel exterior de visualizacion
    {
        lectura =archivo.read();       //el metodo read devuelve 1 caracter cada vez
        estado[i]=lectura;
        if(i==7)       //este if es para que cuando se llegue al final de la linea del fichero de texto, lea el salto de linea y quede preparado para la lectura del siguiente estado
        {
            archivo.read();
        }
    }
}
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

void mostrar_estado()                                                                                                                                                                                               //  MODULO  ACABADO
{
    for (int i = 0 ; i <= 7; i++)
             {
                 if(estado[i] == '1')
                    digitalWrite(panel_led[i], HIGH);
                 if(estado[i] == '0')
                    digitalWrite(panel_led[i], LOW);
             }
}
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

//Este modulo se encarga de hacer la lectura COMPLETA de una línea de un fichero cuya direccion le llega como dato                                                     
String leer_linea(String direccion_archivo)                                                                                                                                                                         //  MODULO  ACABADO
{
    archivo.close();        //cerramos el archivo por sie estaba abierto

    String linea_leida;
    char caracter;

    archivo = SD.open(direccion_archivo, FILE_READ);
    //Serial.println("Existe el archivo: " + String(direccion_archivo));
    if(archivo)     //en caso de que exista y se pueda abrir el archivo...
     {
         //Serial.println("Hemos logrado abrir el archivo: " + String(direccion_archivo));
         while (archivo.available())        //mientras el archivo esté disponible (es decir, no hayamos llegado a su final)...
         {
             caracter = archivo.read();
             linea_leida = linea_leida + caracter;
             //Serial.println("Lo leido es: " + String(linea_leida));
         }
        archivo.close();        //cerramos el archivo
     }
     else       //en caso de que falle la apertura del archivo...
        Serial.println("Error de apertura del archivo: " + String(direccion_archivo));

    return linea_leida;
}

//*************************************************************************************************************************************************************************************************************************************
//*******************************************************************     MODULOS  QUE  HACEN  USO  DE  DATOS  DE  LA  TARGETA  SD      ***********************************************************************************************
//*************************************************************************************************************************************************************************************************************************************


/*Este modulo servirá para tomar el tiempo entre estados que se ha elegido en el ciclo actual y poder decirselo posteriormente al cliente (maestro)*/
int leer_tiempo()                                                                                                                                                                                                   //  MODULO  ACABADO
{
    int tiempo = 0;
    String string_tiempo;

    string_tiempo = leer_linea("/tiempo.txt");

    tiempo = string_tiempo.toInt();     // conversión de la cadena (string) a int
    //tiempo = tiempo * 1000;             // conversión de segundos a milisegundos
    Serial.println("El tiempo entre estados es: " + String(tiempo));

    return tiempo;
}
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

/*Este modulo servirá para tomar el nombre del ciclo que se va a ejecutar y poder decirselo posteriormente al cliente (maestro)*/
String leer_nombre_ciclo()                                                                                                                                                                                          //  MODULO  ACABADO
{
    String nombre = "";

    nombre = leer_linea("/nombre.txt");
    //Serial.println("El nombre del ciclo de trabajo en curso es: " + String(nombre));

    return nombre;
}
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

/*Este modulo se encarga de realizar la getion de los datos de la tarjeta SD desde las lecturas de los ficheros donde obtener los datos del trabajo actual, 
así como de escribirlos en las caracteristicas correspondientes para que el dispositivo maestro tenga constancia de ello */ 
int tomar_tiempo_y_nombre()                                                                                                                                                                                        //  MODULO  ACABADO
{
    String nombre_ciclo = "";
    int tiempo_entre_estados;

    //Serial.println("Hemos empezado a cambiar el nombre y tiempo en el core " );

    // Gestion de la lectura de los nombres del ciclo de trabajo y el tiempo entre estados
    nombre_ciclo = leer_nombre_ciclo();
    escribir_ciclo_en_curso(nombre_ciclo);
    tiempo_entre_estados = leer_tiempo();
    escribir_tiempo_ciclo(tiempo_entre_estados);

    Serial.println( "Hemos actualizado el tiempo y el nombre del ciclo \n");

    return tiempo_entre_estados;
}


//*************************************************************************************************************************************************************************************************************************************
//*********************************************************     MODULOS  REFERENTES  A  LA  ACTUALIZACION  DE LOS  ARCHIVOS  DE  TEXTO      *******************************************************************************************
//*************************************************************************************************************************************************************************************************************************************

void actualizar_ciclo()                                                                                                                                                                                             //  MODULO  ACABADO
{
    archivo.close();        //cerramos los archivos por si estaban abiertos
    nuevo.close();

    char caracter;

    SD.remove("/ciclo.txt"); // borramos el fichero
    
    archivo = SD.open("/ciclo.txt", FILE_WRITE);    //abrimos el archivo que vamos a escribir, sabiendo que en caso de no existir, con esta declaracion, se crea y luego se puede escribir en el, todo dentro de la tarjeta SD
    nuevo = SD.open("/nuevo_ciclo.txt", FILE_READ); // abrimos el archivo desde el que vamos a caragar los datos
    if(nuevo)     //en caso de que exista y se pueda abrir el archivo...
    {
        Serial.println("Procedemos a actualizar la informacion en el fichero del ciclo  de trabajo en el core ");
        while (nuevo.available())        //mientras el archivo esté disponible (es decir, no hayamos llegado a su final)...
        {
            caracter = nuevo.read();
            archivo.print(caracter);
            //Serial.println("Lo leido es: " + String(linea_leida));
        }
        nuevo.close();        //cerramos los archivos
        archivo.close();
    }
    ciclo_actualizado = 1;  // Finalmente, indicamos que ya está actualizado el ciclo de trabajo
}
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

void actualizar_nombre()                                                                                                                                                                                            //  MODULO  ACABADO
{
    archivo.close();        //cerramos los archivos por si estaban abiertos
    nuevo.close();

    char caracter;

    SD.remove("/nombre.txt");

    archivo = SD.open("/nombre.txt", FILE_WRITE);
    nuevo = SD.open("/nuevo_nombre.txt", FILE_READ);// con esta declaración, en caso de que no exista el fichero, se crea y luego escribe en él, todo ello dentro de la tarjeta SD
    if(nuevo)     //en caso de que exista y se pueda abrir el archivo...
      {
          //Serial.println("Procedemos a actualizar la informacion en el fichero del nombre del ciclo de trabajo en el core ");
          while (nuevo.available())        //mientras el archivo esté disponible (es decir, no hayamos llegado a su final)...
          {
                caracter = nuevo.read();
                archivo.print(caracter);
                //Serial.println("Lo leido es: " + String(linea_leida));
          }
          nuevo.close();        //cerramos los archivos
          archivo.close();
      }
    nombre_actualizado = 1;
}
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

void actualizar_tiempo()                                                                                                                                                                                            //  MODULO  ACABADO
{
    archivo.close();        //cerramos los archivos por si estaban abiertos
    nuevo.close();

    char caracter;

    SD.remove("/tiempo.txt");

    archivo = SD.open("/tiempo.txt", FILE_WRITE);
    nuevo = SD.open("/nuevo_tiempo.txt", FILE_READ);// con esta declaración, en caso de que no exista el fichero, se crea y luego escribe en él, todo ello dentro de la tarjeta SD
    if(nuevo)     //en caso de que exista y se pueda abrir el archivo...
    {
        Serial.println("Procedemos a actualizar la informacion en el fichero del tiempo entre estados en el core ");
        while (nuevo.available())        //mientras el archivo esté disponible (es decir, no hayamos llegado a su final)...
        {
            caracter = nuevo.read();
            archivo.print(caracter);
            //Serial.println("Lo leido es: " + String(linea_leida));
        }
        nuevo.close();        //cerramos los archivos
        archivo.close();
    }
    tiempo_actualizado = 1;
}


//*************************************************************************************************************************************************************************************************************************************
//**************************************************     MODULOS  REFERENTES  A  LA  ESCRITURA  DE  FICHEROS  DESDE  LOS  DATOS  RECIBIDOS      ***************************************************************************************
//*************************************************************************************************************************************************************************************************************************************

void escribir_ciclo(std::string ciclo)                                                                                                                                                                              //  MODULO  ACABADO
{
    escribiendo_ciclo = 1;

    nuevo.close();  // por si estaba abierto, cierro este archivo

    Serial.println ("\n\n=== Empezamos a gestionar el fichero ===");
    SD.remove("/nuevo_ciclo.txt");
    nuevo = SD.open("/nuevo_ciclo.txt", FILE_WRITE);// con esta declaración, en caso de que no exista el fichero, se crea y luego escribe en él, todo ello dentro de la tarjeta SD
    int i=0;
    String caracter;
    for(int j = 0 ; j <= ciclo.length() ; j ++)
    {
        caracter = ciclo[j];
        nuevo.print(caracter);        
        if(j == (7 + (8*i)) )
        {
            nuevo.print("\n");
            i = i + 1;
        }
    }
    nuevo.close();

    escribiendo_ciclo = 0;

    Serial.println("***** Hemos acabado de trabajar con el fichero ********\n\n");
}
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

void escribir_nombre_tiempo(std::string dato)                                                                                                                                                                       //  MODULO  ACABADO
{
    escribiendo_tiempo_nombre = 1;

    nuevo.close();  // por si estaba abierto, cierro este archivo

    String caracter;

    /*Se prevee un tamaño MAXIMO en segundos correspondiente a 4 cifras, de modo que pueda haber un tiempo maximo entre estados de hasta 9999 segundos = 2h + 46min + 39 seg, de modo que los datos recibidos SIEMPRE seguiran un orden
    determinado:
        -> Primeros 4 caracteres: se corresponderán con el tiempo
        -> Resto cracteres: se corresponderán con el  nombre del ciclo
    Teniendo esto en cuenta, asaremos a trtar los datos de la manera necesaria para su uso, de modo que primero gestionamos el tiempo...*/    
    SD.remove("/nuevo_tiempo.txt");
    nuevo = SD.open("/nuevo_tiempo.txt", FILE_WRITE);// con esta declaración, en caso de que no exista el fichero, se crea y luego escribe en él, todo ello dentro de la tarjeta SD
    for(int i = 0; i <= 4; i++)
    {
        caracter = dato[i];
        nuevo.print(caracter);
    }
    nuevo.close();

    //... y a continuación, gestionamos el nombre
    if(SD.remove("/nuevo_nombre.txt"))
    {
      Serial.println("Se ha borrado el fichero nuevo_nombre_ciclo para poder escribir el nuevo");
    }
    nuevo = SD.open("/nuevo_nombre.txt", FILE_WRITE);// con esta declaración, en caso de que no exista el fichero, se crea y luego escribe en él, todo ello dentro de la tarjeta SD
    for(int i = 5; i< dato.length(); i++)
    {
        caracter = dato[i];
        nuevo.print(caracter);
    }
    nuevo.close();

    escribiendo_tiempo_nombre = 0;
}


/*-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
****************************************************************************    CLASES  Y  DEFINICIONES  DEL  BLE    ******************************************************************************************************************
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

class MyServerCallbacks : public BLEServerCallbacks                                                                                                                                                                 //  MODULO  ACABADO
{
    //este void comprobará si está conectado el BLE del dispositivo
    void onConnect(BLEServer* mi_servidor)
    {
        Serial.println("Servidor ocupado");
        deviceConnected = true;
    };
    //este void comprobará si NO está conectado el BLE del dispositivo
    void onDisconnect(BLEServer* mi_servidor)
    {
        Serial.println("Servidor vacio");
        deviceConnected = false;
    };
};
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

class MyCallbacks_nuevo_ciclo : public BLECharacteristicCallbacks                                                                                                                                                   //  MODULO  ACABADO
{
     void onWrite (BLECharacteristic *caracteristica_recibo_ciclo)
    {
        std::string ciclo = caracteristica_recibo_ciclo -> getValue();
        if(ciclo.length()>0)
        {
            Serial.println ("=== Empezamos a recibir ===");
            /*
            for(int i = 0; i < tam ; i++)
            {
                Serial.print(ciclo[i]);
            }

            Serial.println();
            
            String recibido ="";
            for(int i = 0 ; i < ciclo.length() ; i++ )
            {
                recibido = recibido + ciclo[i];
            }
            
            delay(1000);
            Serial.println("Se ha comprobado que la caracteristica recibida es la correspondiente a recibir ciclo de trabajo");
            Serial.print("lo recibido en el callback es: ");
            Serial.println(String(recibido));
            */
            
            escribir_ciclo(ciclo);
            ciclo_actualizado = 0;

            Serial.println("==== Finaliza el callback ===");
        }
    }        
};
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

class MyCallbacks_nuevo_nombre : public BLECharacteristicCallbacks                                                                                                                                                  //  MODULO  ACABADO
{
    void onWrite (BLECharacteristic *caracteristica_recibo_nombre_tiempo)
    {
        std::string datos = caracteristica_recibo_nombre_tiempo -> getValue();
        Serial.println ("=== Empezamos a recibir NOMBRE Y TIEMPO ===");
        /*
        Serial.println("Se ha comprobado que la caracteristica recibida es la correspondiente a recibir nombre del ciclo de trabajo");
        
        Serial.print("lo recibido es: ");
        Serial.print(String(recibido));
        delay(1000);
        */
        escribir_nombre_tiempo(datos);
        /*
         if(SD.remove("nuevo_nombre.txt"))
        {
        Serial.println("Se ha borrado el fichero nuevo_nombre_ciclo para poder escribir el nuevo");
        }
        delay(1000);
        String string_nombre;
        nuevo = SD.open("/nuevo_nombre.txt", FILE_WRITE);// con esta declaración, en caso de que no exista el fichero, se crea y luego escribe en él, todo ello dentro de la tarjeta SD
        for(int i = 0; i< rcvString.length(); i++)
        {
            string_nombre = string_nombre + rcvString[i];
        }
        nuevo.close();
        nuevo.print(string_nombre);
        nuevo.println();
        */
       tiempo_actualizado = 0;
       nombre_actualizado = 0;
       Serial.println("==== Finaliza el callback NOMBRE Y TIEMPO ===");
    }
};


/*-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
***************************************************************************************************   MODULOS    **********************************************************************************************************************
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/


//*************************************************************************************************************************************************************************************************************************************
//*********************************************************     MODULOS  REFERENTES  A  LA  INICIALIZACION  DEL  DISPOSITIVO  ESCLAVO      ********************************************************************************************
//*************************************************************************************************************************************************************************************************************************************

void inicializacion_hardware()                                                                                                                                                                                      //  MODULO  ACABADO
{
    //inicializo las salidas de control desde el ESP32, es decir, las señales al panel led
    for (int i = 0; i <= 7; i++)
    {
        pinMode(panel_led [i], OUTPUT);
        digitalWrite(panel_led [i], LOW);
    }

    //inicializo los valores de los pines que controlan el led RGB
    for (int i = 0; i <= 2; i++)
    {
        pinMode(led_rgb[i], OUTPUT);
        analogWrite(led_rgb[i], 0);
    }

    //compruebo el modulo lector tarjetas SD
    if(!SD.begin(CS))
    {
      Serial.println("Fallo inicializacion lector micro SD");
      //ahora, conviene mandar un mensaje de error al maestro para que sepa el fallo ocurrido
       return;
    }
    Serial.println("Inicializacion correcta lector micro SD");
    //ahora, conviene mandar un mensaje al maestro para que sepa la situacion del hardware.
}
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

void inicializar_BLE()                                                                                                                                                                                              //  MODULO  ACABADO
{
    //EL ORDEN EN CUANTO A SERVIDOR, SERVICIOS Y CARACTERÍSTICAS ES OBLIGATORIAMENTE EL AQUI SEGUIDO
    Serial.println("Estamos inicializando el servidor");

    //Creo el dispositivo BLE y le doy nombre
    BLEDevice::init("BLE ESP32 panel luminoso");

    //Creo el servidor BLE y un objeto de la clase MyServerCallbacks para atender a los accesos a dicho servidor. Hemos de saber que servidor y perfil (server and profile) se usan indistintamente en nomenclatura inglesa
    mi_servidor = BLEDevice::createServer();
    mi_servidor -> setCallbacks(new MyServerCallbacks);

    //Creo el servicio BLE y se lo añado al servidor creado. Tambien le asigno el UUID definido a traves de "SERVICE_UUID", para acceder a el
    servicio_transferencia = mi_servidor -> createService(SERVICE_UUID);

    //Creo las características del BLE, pues será a través de ellas como definiremos la comunicacion con el smartphone (maestro)
    caracteristica_envio_estado_actual = servicio_transferencia -> createCharacteristic( CHARACTERISTIC_UUID_ESTADO, BLECharacteristic::PROPERTY_NOTIFY | BLECharacteristic::PROPERTY_READ);
    caracteristica_envio_tiempo = servicio_transferencia -> createCharacteristic(CHARACTERISTIC_UUID_TIEMPO, BLECharacteristic::PROPERTY_NOTIFY | BLECharacteristic::PROPERTY_READ);
    caracteristica_envio_nombre_ciclo = servicio_transferencia -> createCharacteristic(CHARACTERISTIC_UUID_CICLO, BLECharacteristic::PROPERTY_NOTIFY | BLECharacteristic::PROPERTY_READ);

    caracteristica_recibo_ciclo = servicio_transferencia -> createCharacteristic(CHARACTERISTIC_UUID_RECIBE_CICLO , BLECharacteristic::PROPERTY_WRITE);
    caracteristica_recibo_nombre_tiempo = servicio_transferencia -> createCharacteristic(CHARACTERISTIC_UUID_RECIBE_NOMBRE_TIEMPO , BLECharacteristic::PROPERTY_WRITE);

    //Preparo las llamadas para escritura en las caracteristicas con dicha propiedad
    caracteristica_recibo_ciclo -> setCallbacks(new MyCallbacks_nuevo_ciclo);
    caracteristica_recibo_nombre_tiempo -> setCallbacks(new MyCallbacks_nuevo_nombre);

    //BL2902 necesidad de notificar
    caracteristica_envio_estado_actual -> addDescriptor(new BLE2902());
    caracteristica_envio_tiempo -> addDescriptor(new BLE2902());
    caracteristica_envio_nombre_ciclo -> addDescriptor(new BLE2902());

    //Arrancamos el servicio y lo notificamos al servidor
    servicio_transferencia -> start();
    mi_servidor -> getAdvertising() -> start();

    //Notificamos por pantalla que esto se ha realizado correctamente
    Serial.println("Esperando una conexión con cliente para notificar...");
}

//*************************************************************************************************************************************************************************************************************************************
//********************************************     MODULOS  REFERENTES  A  GESTIONES  INTERNAS  DEL  ESCLAVO  PARA  SU  CORRECTO  FUNCIONAMIENTO      *********************************************************************************
//*************************************************************************************************************************************************************************************************************************************

/*Este modulo servirá para calcular cuánto delay provoca el programa y así compensar el delay deseado por el dispositivo maestro*/                                                                                  //  MODULO  ACABADO
int compensar_tiempo(int tiempo) //paso la variable "delay programa" por referencia, para que pueda ser modificada
{
    int compensacion;
    //Muestro por pantalla y meto un delay para que haya tiempo suficiente como para ver las diferentes lecturas
    Serial.println("Retraso temporal provocado por la ejecucion del programa: " + delay_programa);
    delay(1000);
    compensacion = tiempo - delay_programa;
    return compensacion;
}


//*************************************************************************************************************************************************************************************************************************************
//*********************************************************     MODULOS  REFERENTES  AL  LED  DE  INDICACION  DE  ESTADO  DE  CONEXION      *******************************************************************************************
//*************************************************************************************************************************************************************************************************************************************

void verde()                                                                                                                                                                                                        //  MODULO  ACABADO
{
    analogWrite(rgb_verde,255);
    analogWrite(rgb_rojo,0);
    analogWrite(rgb_azul,0);
}
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

void amarillo()                                                                                                                                                                                                     //  MODULO  ACABADO
{
    analogWrite(rgb_verde,80);
    analogWrite(rgb_rojo,255);
    analogWrite(rgb_azul,0);
}
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

void rojo()                                                                                                                                                                                                         //  MODULO  ACABADO
{
    analogWrite(rgb_verde,0);
    analogWrite(rgb_rojo,255);
    analogWrite(rgb_azul,0);
}
//_____________________________________________________________________________________________________________________________________________________________________________________________________________________________________

void gestion_led_rgb(int estado_conexion)                                                                                                                                                                           //  MODULO  ACABADO
{
    switch (estado_conexion)
    {
        case 1:  // trabajo con BLE activado Y CON CONEXION a dispositivo maestro
            verde();
            break;
        case 3:     // trabajo con BLE activado pero SIN CONECTAR a dispositivo maestro
            amarillo();
        default:    // trabajo con BLE activado pero SIN CONECTAR a dispositivo maestro
            amarillo();
    }
    Serial.println("Se ha gestionado el color del led indicador de estado \n");
}

//*************************************************************************************************************************************************************************************************************************************
//************************************************************     MODULOS  REFERENTES  A  GESTION  DEL  ESTADO  DE  CONEXION  DEL  SERVIDOR      *************************************************************************************
//*************************************************************************************************************************************************************************************************************************************

void estado_conexion()
{
    if(deviceConnected)
    {
        color = 1;        //será un valor con el que decidiremos el color del led RGB que se utilizará para indicar el estado del hardware esclavo
    }            
    else
    {
        color = 3;        //será un valor con el que decidiremos el color del led RGB que se utilizará para indicar el estado del hardware esclavo
    }
    Serial.println("Se ha comprobado el estado de conexion del dispositivo\n");
    gestion_led_rgb(color);
}


//*************************************************************************************************************************************************************************************************************************************
//************************************************************     MODULOS  REFERENTES  A  GESTION  DEL  CICLO  DE  TRABAJO  EN  CURSO      *******************************************************************************************
//*************************************************************************************************************************************************************************************************************************************

/*Este modulo servirá para tomar el estado del ciclo que se va a mostrar y poder decirselo posteriormente al cliente (maestro)*/
void ciclo_trabajo()                                                                                                                                                                                                //  MODULO  ACABADO
{
    int tiempo;
    //Serial.println("Hemos entrado en el ciclo de trabajo con el core ");
    if(ciclo_actualizado == 1 && tiempo_actualizado == 1 && nombre_actualizado == 1)     //en caso de que exista y se pueda abrir el archivo...
     {
         tiempo = tomar_tiempo_y_nombre();   // de esta manera, actualizamos los valores del tiempo y del nombre en el core ue se encarga de gestionar el panel, en caso de que los ficheros se hayan actualizado por el otro core

         archivo = SD.open("/ciclo.txt");
         Serial.println("Se ha abierto el archivo ciclo.txt");
         while (archivo.available() && ciclo_actualizado == 1 && tiempo_actualizado == 1 && nombre_actualizado == 1)        //mientras el archivo esté disponible (es decir, no hayamos llegado a su final), Y el archivo sea el actual ...
         {
             estado_conexion();
             //Serial.println("Entramos en el bucle While");
             cargar_estado();
             mostrar_estado();
             escribir_estado();
             delay(tiempo);
         }
        archivo.close();        //cerramos el archivo
     }
     else
     {
         delay(500); // este delay es para que haya tiempo para manejar los ficheros por parte del callback
         if(escribiendo_ciclo == 0 && escribiendo_tiempo_nombre == 0)
         {
             actualizar_ciclo();
             actualizar_nombre();
             actualizar_tiempo();
         }        
     }
}


/*-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
*************************************************************************************************************   SETUP    **************************************************************************************************************
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

void setup()                                                                                                                                                                                                        //  MODULO  ACABADO
{
  Serial.begin(9600);  
  inicializacion_hardware();
  inicializar_BLE();
  //Serial.println("La inicializacion del sistema ha sido correcta");
}

/*-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
**************************************************************************************************************   LOOP    **************************************************************************************************************
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

void loop()                                                                                                                                                                                                         //  MODULO  ACABADO
{
    ciclo_trabajo();
}