package com.example.tfg;

import androidx.appcompat.app.AppCompatActivity;
import androidx.core.app.ActivityCompat;
import androidx.core.content.ContextCompat;

import android.Manifest;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCallback;
import android.bluetooth.BluetoothManager;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageManager;
import android.graphics.Color;
import android.net.Uri;
import android.os.Bundle;
import android.os.Handler;
import android.util.Log;
import android.view.View;
import android.view.accessibility.AccessibilityNodeInfo;
import android.widget.Button;
import android.widget.ImageButton;
import android.widget.ImageView;
import android.widget.LinearLayout;
import android.widget.ListView;
import android.widget.TextView;

import java.util.ArrayList;
import java.util.HashMap;

public class Accessible_Devices extends AppCompatActivity implements View.OnClickListener
{
    private static final String TAG = Accessible_Devices.class.getName();

    //Variables referentes a los objetos visibles en el layout
    Button find_devices;
    ImageButton help_ad;
    LinearLayout ll_accesible_devices;

    //Variables referentes a la gestion del BLE
    BluetoothManager bluetoothManager;
    BluetoothAdapter bluetoothAdapter;
    BroadcastReceiver broadcastReceiver;

    // Variable para acceder a los diferentes dispositivos BLE accesibles
    ArrayList<BluetoothDevice> devices;

    /*_++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++_*/
    /*__________________________________________________________________ BASE  METHODS  OF  THE  ACTIVITY ______________________________________________________________________________________*/
    /*_++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++_*/

    @Override
    protected void onCreate(Bundle savedInstanceState)
    {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.accessible_devices);
        checkPermissions();
        init();
    }
    /*_________________________________________________________________________________________________________________________________________________________________________________________*/

    @Override
    public void onClick(View view)                                                                                                                                                  //  ACABADO
    {
        switch (view.getId())
        {
            case R.id.bt_find_devices_ad:
                bluetoothAdapter.cancelDiscovery();
                devices.clear();    // esta linea y la anterior han sido intercambiadas de posición entre sí a modo de prueba
                ll_accesible_devices.removeAllViewsInLayout();
                find_devices();
                break;
            case R.id.bt_help_ad:
                help_of_this_activity();
                break;
            default:
                break;
        }
    }
    /*_________________________________________________________________________________________________________________________________________________________________________________________*/

    private void find_devices()
    {
        Log.i(TAG, "Buscando dispositivos ");
        //Para poder encontrar dispositivos, necesito crear un elemento de escucha y editar las acciones que realizará
        broadcastReceiver = new BroadcastReceiver()
        {
            @Override
            public void onReceive(Context context, Intent intent)
            {
                //el elemento de escucha puede recibir distintos tipos de avisos, así que tenemos que discernir cuando el aviso significa "dispositivo descubierto"
                String action = intent.getAction();
                if (action.equals(BluetoothDevice.ACTION_FOUND)) //si lo escuchado es un intento de encontrar...
                {
                    final BluetoothDevice device = intent.getParcelableExtra(BluetoothDevice.EXTRA_DEVICE); //creamos un dispositivo bluetooth (una variable de acceso) tomando sus datos (get extra)y guardandolos
                    devices.add(device);    //lo añadimos a la lista
                    Log.i(TAG, "Dispositivo descubierto: " + device.getAddress());  // linea de depuración, que nos da por el logcat (que hace las veces de historial en tiempo real) la direccion del dispositivo bluetooth encontrado

                    // creamos una vista para cada uno de los elementos (item) encontrados, bajo la forma definida en item_lv_view_device. para ello, primero definimos los elementos de ese layout
                    final View view = View.inflate(Accessible_Devices.this, R.layout.item_lv_view_device, null);
                    TextView name = view.findViewById(R.id.name_device_lv_vd);
                    ImageView image = view.findViewById(R.id.state_connection_lv_vd);

                    //editamos los valores del item en el linear layout con los que acabamos de encontrar en el dispositivo bluetooth que se ha encontrado
                    name.setText(device.getName());
                    image.setImageResource(R.color.disconnected);

                    //definimos la acciones a realizar cuando hagamos click en un elemento del linear layout
                    view.setOnClickListener(new View.OnClickListener()
                    {
                        @Override
                        public void onClick(View view)
                        {
                            onClickElement(device);
                        }
                    });

                    //añadimos la vista creada al linear layout como un nuevo elemento
                    runOnUiThread(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            ll_accesible_devices.addView(view);
                        }
                    });
                    Log.i(TAG , " Hemos añadido otro elemento al linear layout, y su direccion es " + device.getAddress());
                }
            }
        };

        //filtramos los valores de los dispositivos escuchados encontrados
        IntentFilter filter = new IntentFilter(BluetoothDevice.ACTION_FOUND);
        registerReceiver(broadcastReceiver, filter);

        Log.i(TAG, "BroadcastReceiver registrado");

        // startDiscovery es un método que inicializa la búsqueda de dispositivos bluetooth
        bluetoothAdapter.startDiscovery();
        Log.i(TAG, "startDiscovery()");

        // creamos un manejador de tiempo para limitar el tiempo de uso del escaner bluettoth con el fin de evitar que esté continuamente escaneando (por su elevado consumo de energía)
        Handler handler = new Handler();
        handler.postDelayed(new Runnable()
        {
            @Override
            public void run()
            {
                bluetoothAdapter.cancelDiscovery();
                Log.i(TAG, "cancelDiscovery()");
            }
        }, 10 * 1000);
    }
    /*_________________________________________________________________________________________________________________________________________________________________________________________*/

    public void checkPermissions()
    {
        //TODO poner el metodo
    }

    /*_++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++_*/
    /*__________________________________________________________________________ CONTROL  THE  BUTTONS _________________________________________________________________________________________*/
    /*_++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++_*/

    private void onClickElement(BluetoothDevice device)                                                                                                                             //  ACABADO
    {

        unregisterReceiver(broadcastReceiver);
        bluetoothAdapter.cancelDiscovery();
        Intent activity = new Intent(Accessible_Devices.this, Device_View.class);   // creamos un tunel de acceso a la activity "Device_View"
        activity.putExtra("DEVICE", device);    // ponemos en común los datos del dispositivo bluetooth seleccionado
        startActivityForResult(activity, 1122);    //arrancamos la otra activity
    }
    /*_________________________________________________________________________________________________________________________________________________________________________________________*/

    private void help_of_this_activity()
    {
        //AMBOS CODIGOS SON VALIDOS, PERO VAMOS A USAR EL INFERIOR PARA EVITAR TENER OTRA VARIABLE EN JUEGO

        // Intent youtube_help = new Intent(Intent.ACTION_VIEW, Uri.parse("https://www.youtube.com/watch?v=9-3OCc5g5oE"));//desde la ventana en que estoy, a la que quiero llegar
        // startActivity(youtube_help);
        startActivity(new Intent(Intent.ACTION_VIEW, Uri.parse("https://www.youtube.com/watch?v=9-3OCc5g5oE")));
        // Toast.makeText(this, "Hemos pasado por la parte de ir a youtube",Toast.LENGTH_SHORT).show(); // Los "toast" que uso son solo para verificacion de por donde llega el codigo en ejecucion
    }

    /*_++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++_*/
    /*___________________________________________________________________________ METODH  ON  RESULT ___________________________________________________________________________________________*/
    /*_++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++_*/

    @Override
    public void onActivityResult(int requestCode, int resultCode, Intent data)
    {
        super.onActivityResult(requestCode, resultCode, data);
        Log.i(TAG, "onActivityResult");
        switch (requestCode)
        {
            case 1122:
                Log.i(TAG, "onActivityResult --> Volvemos desde ver un dispositivo al que habiamos accedido");
                switch (resultCode)
                {
                    case RESULT_OK:
                        bluetoothAdapter.cancelDiscovery();// LINEA NUEVA !
                        devices.clear();
                        ll_accesible_devices.removeAllViewsInLayout();
                        find_devices();
                        break;
                    case RESULT_CANCELED:
                        Log.i(TAG, "No se ha decidido configurar un nuevo trabajo");
                }
        }
    }

    /*_++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++_*/
    /*_________________________________________________________________________ PREPARE  THE  ACTIVITY _________________________________________________________________________________________*/
    /*_++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++_*/

    private void prepare_buttons()                                                                                                                                                    //  ACABADO
    {
        Log.i(TAG, "Preparamos los botones");
        find_devices.setOnClickListener(this);
        help_ad.setOnClickListener(this);
    }
    /*_________________________________________________________________________________________________________________________________________________________________________________________*/


    private void init()                                                                                                                                                               //  ACABADO
    {
        // Elementos de la activity
        find_devices = findViewById(R.id.bt_find_devices_ad);
        help_ad = findViewById(R.id.bt_help_ad);
        ll_accesible_devices = findViewById(R.id.ll_devices_ad);

        // Creo una variable del tipo bluetoothmanager pidiendo servicio bluetooth al dispositivo maestro y luego pido su adaptador
        bluetoothManager = (BluetoothManager) getSystemService(BLUETOOTH_SERVICE);
        bluetoothAdapter = bluetoothManager.getAdapter();

        prepare_buttons();

        //inicializacion del arraylist
        devices = new ArrayList<>();

        //hacemos una primera búsqueda de dispositivos, sin necesidad de pulsar el boton "buscar dispositivos"
        find_devices();
    }

}
